@@ -1,6 +1,6 @@
 package Juju;
-$Juju::VERSION = '1.0';
-# ABSTRACT: Non-blocking Perl bindings for http://juju.ubuntu.com/
+$Juju::VERSION = '1.5.1';
+# ABSTRACT: Perl bindings for Juju
 
 
 use strict;
@@ -18,20 +18,16 @@ __END__
 
 =head1 NAME
 
-Juju - Non-blocking Perl bindings for http://juju.ubuntu.com/
+Juju - Perl bindings for Juju
 
 =head1 VERSION
 
-version 1.0
+version 1.5.1
 
 =head1 DESCRIPTION
 
-Pure perl non-blocking bindings for Juju. See L<Juju::Manual::Quickstart> for more information.
-
-=head1 NOTICE
-
-Code is B<alpha quality> and subject to change. API breakage will occur
-until this project reaches 1.x version.
+Perl non-blocking/blocking bindings for Juju. See
+L<Juju::Manual::Quickstart> for more information.
 
 =head1 AUTHOR
 
@@ -45,6 +41,18 @@ This is free software, licensed under:
 
   The MIT (X11) License
 
+=head1 SEE ALSO
+
+Please see those modules/websites for more information related to this module.
+
+=over 4
+
+=item *
+
+L<https://juju.ubuntu.com|https://juju.ubuntu.com>
+
+=back
+
 =head1 DISCLAIMER OF WARRANTY
 
 BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
