@@ -1,13 +1,15 @@
 package Juju::Environment;
-$Juju::Environment::VERSION = '1.0';
+$Juju::Environment::VERSION = '1.5.1';
 # ABSTRACT: Exposed juju api environment
 
 
 use strict;
 use warnings;
-use HTTP::Tiny;
 use JSON::PP;
-use Devel::Deprecate 'deprecate';
+use YAML::Tiny qw(Dump);
+use Data::Validate::Type qw(:boolean_tests);
+use Params::Validate qw(:all);
+use Juju::Util;
 use parent 'Juju::RPC';
 
 
@@ -15,32 +17,16 @@ use Class::Tiny qw(password is_authenticated), {
     endpoint => sub {'wss://localhost:17070'},
     username => sub {'user-admin'},
     Jobs     => sub {
-        +{  HostUnits     => 'JobHostUnits',
+        {   HostUnits     => 'JobHostUnits',
             ManageEnviron => 'JobManageEnviron',
             ManageState   => 'JobManageSate'
         };
-    }
+    },
+    util   => Juju::Util->new
 };
 
 
 
-sub query_cs {
-    my ($self,   $charm)  = @_;
-    my ($series, $_charm) = $charm =~ /^(precise|trusty)\/(\w+)/i;
-    my $cs_url = 'https://manage.jujucharms.com/api/3/charm';
-    if (!$series) {
-        $series = 'trusty';
-        $_charm = $charm;
-    }
-
-    my $composed_url = sprintf("%s/%s/%s", $cs_url, $series, $_charm);
-    my $res = HTTP::Tiny->new->get($composed_url);
-    die "Unable to query charm store\n" unless $res->{success};
-    return decode_json($res->{content});
-}
-
-
-
 sub _prepare_constraints {
     my ($self, $constraints) = @_;
     foreach my $key (keys %{$constraints}) {
@@ -54,48 +40,36 @@ sub _prepare_constraints {
 
 sub login {
     my $self = shift;
-    $self->create_connection unless $self->is_connected;
-    $self->call(
-        {   "Type"      => "Admin",
-            "Request"   => "Login",
-            "RequestId" => $self->request_id,
-            "Params"    => {
-                "AuthTag"  => $self->username,
-                "Password" => $self->password
-            }
-        },
-        sub {
-            $self->is_authenticated(1);
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+
+    my $params = {
+        "Type"      => "Admin",
+        "Request"   => "Login",
+        "RequestId" => $self->request_id,
+        "Params"    => {
+            "AuthTag"  => $self->username,
+            "Password" => $self->password
         }
-    );
+    };
+
+    # block
+    my $res = $self->call($params);
+    die $res->{Error} if defined($res->{Error});
+    $self->is_authenticated(1)
+      unless !defined($res->{Response}->{EnvironTag});
 }
 
 
 
+
 sub reconnect {
     my $self = shift;
     $self->close;
-    $self->create_connection;
     $self->login;
     $self->request_id = 1;
 }
 
 
-sub info {
-    my $self = shift;
-    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
-    my $params = {"Type" => "Client", "Request" => "EnvironmentInfo"};
-
-    deprecate(
-              reason => 'Please use environment_info() for getting environment',
-              die => '2014-12-01'
-             );
-
-    return $self->environment_info unless $cb;
-    return $self->environment_info($cb);
-}
-
-
 sub environment_info {
     my $self = shift;
     my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
@@ -108,6 +82,82 @@ sub environment_info {
     return $self->call($params, $cb);
 }
 
+sub environment_uuid {
+    my $self = shift;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {"Type" => "Client", "Request" => "EnvironmentUUID"};
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+}
+
+
+
+sub environment_unset {
+    my ($self, $items) = @_;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "EnvironmentUnset",
+        "Params"  => $items
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+}
+
+sub find_tools {
+    my ($self, $major, $minor, $series, $arch) = @_;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "EnvironmentUnset",
+        "Params"  => {
+            MajorVersion => int($major),
+            MinorVersion => int($minor),
+            Arch         => $arch,
+            Series       => $series
+        }
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+}
+
+
+sub agent_version {
+    my $self = shift;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {"Type" => "Client", "Request" => "AgentVersion"};
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+}
+
+sub abort_current_upgrade {
+    my $self = shift;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {"Type" => "Client", "Request" => "AbortCurrentUpgrade"};
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+}
+
 
 
 sub status {
@@ -125,6 +175,21 @@ sub status {
     return $self->call($params, $cb);
 }
 
+sub client_api_host_ports {
+    my $self   = shift;
+    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"   => "Client",
+        "Request" => "APIHostPorts"
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+}
+
 
 
 sub get_watcher {
@@ -188,7 +253,7 @@ sub get_charm {
 }
 
 
-sub get_env_constraints {
+sub get_environment_constraints {
     my $self   = shift;
     my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
     my $params = {
@@ -205,7 +270,7 @@ sub get_env_constraints {
 }
 
 
-sub set_env_constraints {
+sub set_environment_constraints {
     my ($self, $constraints) = @_;
     my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
     my $params = {
@@ -222,7 +287,7 @@ sub set_env_constraints {
 }
 
 
-sub get_env_config {
+sub environment_get {
     my $self   = shift;
     my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
     my $params = {
@@ -238,7 +303,7 @@ sub get_env_config {
 }
 
 
-sub set_env_config {
+sub environment_set {
     my ($self, $config) = @_;
     my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
     my $params = {
@@ -256,19 +321,37 @@ sub set_env_config {
 
 
 sub add_machine {
-    my ($self, $series, $constraints, $machine_spec, $parent_id,
-        $container_type)
-      = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $self = shift;
+    my ($series, $constraints, $machine_spec, $parent_id, $container_type) =
+      validate_pos(
+        @_,
+        {default => 'trusty'},
+        {type    => HASHREF, default => +{}},
+        0, 0, 0
+      );
 
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
     my $params = {
         "Series"        => $series,
-        "Constraints"   => $self->_prepare_constraints($constraints),
-        "ContainerType" => $container_type,
-        "ParentId"      => $parent_id,
-        "Jobs"          => $self->Jobs->{HostUnits},
+        "Jobs"          => [$self->Jobs->{HostUnits}],
+        "ParentId"      => "",
+        "ContainerType" => ""
     };
 
+    # validate constraints
+    if (defined($constraints) and is_hashref($constraints)) {
+        $params->{Constraints} = $self->_prepare_constraints($constraints);
+    }
+
+    # if we're here then assume constraints is good and we can check the
+    # rest of the arguments
+    if (defined($machine_spec)) {
+        die "Cant specify machine spec with container_type/parent_id"
+          if $parent_id or $container_type;
+        ($params->{ParentId}, $params->{ContainerType}) = split /:/,
+          $machine_spec;
+    }
+
     return $self->add_machines([$params]) unless $cb;
     return $self->add_machines([$params], $cb);
 }
@@ -291,6 +374,24 @@ sub add_machines {
 }
 
 
+
+sub destroy_environment {
+    my $self   = shift;
+    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "DestroyEnvironment"
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+
+}
+
+
 sub destroy_machines {
     my ($self, $machine_ids, $force) = @_;
     my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
@@ -349,45 +450,73 @@ sub remove_relation {
 
 
 sub deploy {
-    my ($self, $service_name, $charm_url, $num_units, $config_yaml,
-        $constraints, $machine_spec)
-      = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $self = shift;
+    my ($charm, $service_name) =
+      validate_pos(@_, {type => SCALAR}, {type => SCALAR});
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
 
-    my $params = {ServiceName => $service_name};
+    # parse additional arguments
+    my ($num_units, $config_yaml, $constraints, $machine_spec) = @_;
+
+    my $params = {
+        Type    => "Client",
+        Request => "ServiceDeploy",
+        Params  => {ServiceName => $service_name}
+    };
+
+    # Check for series format
+    my (@charm_args) = $charm =~ /(\w+)\/(\w+)/i;
+    my $_charm_url = undef;
+    if (scalar @charm_args == 2) {
+        $_charm_url = $self->util->query_cs($charm_args[1], $charm_args[0]);
+    }
+    else {
+        $_charm_url = $self->util->query_cs($charm);
+    }
+
+    $params->{Params}->{CharmUrl} = $_charm_url->{charm}->{url};
     $num_units = 1 unless $num_units;
-    $params->{NumUnits}   = $num_units;
-    $params->{ConfigYAML} = $config_yaml;
-    my $svc_constraints;
-    if ($constraints) {
-        $params->{Constraints} = $self->_prepare_constraints($constraints);
+    $params->{Params}->{NumUnits} = $num_units;
+    $params->{Params}->{ConfigYAML} =
+      defined($config_yaml) ? $config_yaml : "";
+
+    if (defined($constraints) and is_hashref($constraints)) {
+        $params->{Params}->{Constraints} =
+          $self->_prepare_constraints($constraints);
     }
     if ($machine_spec) {
-        $params->{ToMachineSpec} = $machine_spec;
+        $params->{Params}->{ToMachineSpec} = "$machine_spec";
     }
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "ServiceDeploy",
-            "Params"  => $params
-        }
-    );
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
-sub set_config {
-    my ($self, $service_name, $config) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+sub service_set {
+    my $self = shift;
+    my ($service_name, $config) =
+      validate_pos(@_, {type => SCALAR}, {type => HASHREF, optional => 1});
 
-    die "Not a hash" unless ref $config eq 'HASH';
-    return $self->call(
-        {   "Type"    => "Client",
-            "Request" => "ServiceSet",
-            "Params"  => {
-                "ServiceName" => $service_name,
-                "Options"     => $config
-            }
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "ServiceSet",
+        "Params"  => {
+            "ServiceName" => $service_name,
+            "Options"     => $config
         }
-    );
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
@@ -395,47 +524,60 @@ sub unset_config {
     my ($self, $service_name, $config_keys) = @_;
     my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
 
-    return $self->call(
+    my $params = 
         {   "Type"    => "Client",
             "Request" => "ServiceUnset",
             "Params"  => {
                 "ServiceName" => $service_name,
                 "Options"     => $config_keys
             }
-        }
-    );
+        };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
 sub set_charm {
     my ($self, $service_name, $charm_url, $force) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
 
     $force = 0 unless $force;
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "ServiceSetCharm",
-            "Params"  => {
-                "ServiceName" => $service_name,
-                "CharmUrl"    => $charm_url,
-                "Force"       => $force
-            }
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "ServiceSetCharm",
+        "Params"  => {
+            "ServiceName" => $service_name,
+            "CharmUrl"    => $charm_url,
+            "Force"       => $force
         }
-    );
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
-sub get_service {
+sub service_get {
     my ($self, $service_name) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "ServiceGet",
+        "Params"  => {"ServiceName" => $service_name}
+    };
 
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "ServiceGet",
-            "Params"  => {"ServiceName" => $service_name}
-        },
-        sub { my $res = shift; return $res }
-    );
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
@@ -443,245 +585,331 @@ sub get_config {
     my ($self, $service_name) = @_;
     my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
 
-    my $svc = $self->get_service($service_name);
-    return $svc->{Config};
+    my $svc = $self->service_get($service_name);
+    return $svc->{Config} unless $cb;
+    return $cb->($svc->{Config});
 }
 
 
-sub get_constraints {
+sub get_service_constraints {
     my ($self, $service_name) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "GetServiceConstraints",
+        "Params"  => {"ServiceName" => $service_name}
+    };
 
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "GetServiceConstraints",
-            "Params"  => {"ServiceName" => $service_name}
-        },
-        sub {
-            my $res = shift;
-            return $res->{Constraints};
-        }
-    );
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
-sub set_constraints {
+sub set_service_constraints {
     my ($self, $service_name, $constraints) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
-
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "SetServiceConstraints",
-            "Params"  => {
-                "ServiceName" => $service_name,
-                "Constraints" => $self->_prepare_constraints($constraints)
-            }
-        },
-        sub { my $res = shift; return $res }
-    );
-}
-
-
-sub update_service {
-    my ($self, $service_name, $charm_url, $force_charm_url,
-        $min_units, $settings, $constraints)
-      = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
-
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "SetServiceConstraints",
-            "Params"  => {
-                "ServiceName"     => $service_name,
-                "CharmUrl"        => $charm_url,
-                "MinUnits"        => $min_units,
-                "SettingsStrings" => $settings,
-                "Constraints"     => $self->_prepare_constraints($constraints)
-            }
-        },
-        sub { my $res = shift; return $res }
-    );
-}
-
-
-sub destroy_service {
-    my ($self, $service_name) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
-
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "ServiceDestroy",
-            "Params"  => {"ServiceName" => $service_name}
-        },
-        sub { my $res = shift; return $res }
-    );
-}
-
-
-sub expose {
-    my ($self, $service_name) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
-
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "ServiceExpose",
-            "Params"  => {"ServiceName" => $service_name}
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "SetServiceConstraints",
+        "Params"  => {
+            "ServiceName" => $service_name,
+            "Constraints" => $self->_prepare_constraints($constraints)
         }
-    );
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+}
+
+sub share_environment {
+    my ($self, $users) = @_;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "ShareEnvironment",
+        "Params"  => $users
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+}
+
+sub unshare_environment {
+    my ($self, $users) = @_;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "UnshareEnvironment",
+        "Params"  => $users
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
-sub unexpose {
+
+sub service_destroy {
     my ($self, $service_name) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "ServiceDestroy",
+        "Params"  => {"ServiceName" => $service_name}
+    };
 
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "ServiceUnexpose",
-            "Params"  => {"ServiceName" => $service_name}
-        },
-        sub { my $res = shift; return $res }
-    );
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
-sub valid_relation_names {
+sub service_expose {
     my ($self, $service_name) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "ServiceExpose",
+        "Params"  => {"ServiceName" => $service_name}
+    };
 
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "ServiceCharmRelations",
-            "Params"  => {"ServiceName" => $service_name}
-        },
-        sub { my $res = shift; return $res }
-    );
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
-sub add_units {
+
+sub service_unexpose {
+    my ($self, $service_name) = @_;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "ServiceUnexpose",
+        "Params"  => {"ServiceName" => $service_name}
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+}
+
+
+sub service_charm_relations {
+    my ($self, $service_name) = @_;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "ServiceCharmRelations",
+        "Params"  => {"ServiceName" => $service_name}
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+}
+
+
+sub add_service_units {
     my ($self, $service_name, $num_units) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
 
     $num_units = 1 unless $num_units;
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "AddServiceUnits",
-            "Params"  => {
-                "ServiceName" => $service_name,
-                "NumUnits"    => $num_units
-            }
-        },
-        sub {
-            my $res = shift;
-            return $res;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "AddServiceUnits",
+        "Params"  => {
+            "ServiceName" => $service_name,
+            "NumUnits"    => $num_units
         }
-    );
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
-sub add_unit {
+sub add_service_unit {
     my ($self, $service_name, $machine_spec) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
 
     $machine_spec = 0 unless $machine_spec;
     my $params = {
-        "ServiceName" => $service_name,
-        "NumUnits"    => 1
+        "Type"    => "Client",
+        "Request" => "AddServiceUnits",
+        "Params"  => {
+            "ServiceName" => $service_name,
+            "NumUnits"    => 1
+        }
     };
 
     if ($machine_spec) {
-        $params->{ToMachineSpec} = $machine_spec;
+        $params->{Params}->{ToMachineSpec} = $machine_spec;
     }
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "AddServiceUnits",
-            "Params"  => $params
-        },
-        sub { my $res = shift; return $res }
-    );
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
-
-sub remove_unit {
+sub destroy_service_units {
     my ($self, $unit_names) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "DestroyServiceUnits",
+        "Params"  => {"UnitNames" => $unit_names}
+    };
 
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "DestroyServiceUnits",
-            "Params"  => {"UnitNames" => $unit_names}
-        },
-        sub { my $res = shift; return $res }
-    );
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
 sub resolved {
     my ($self, $unit_name, $retry) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
 
     $retry = 0 unless $retry;
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "Resolved",
-            "Params"  => {
-                "UnitName" => $unit_name,
-                "Retry"    => $retry
-            }
-        },
-        sub { my $res = shift; return $res }
-    );
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "Resolved",
+        "Params"  => {
+            "UnitName" => $unit_name,
+            "Retry"    => $retry
+        }
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
 
-sub get_public_address {
-    my ($self, $target) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
 
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "PublicAddress",
-            "Params"  => {"Target" => $target}
-        },
-        sub { my $res = shift; return $res; }
-    );
-}
-
-
-sub set_annotation {
+sub set_annotations {
     my ($self, $entity, $entity_type, $annotation) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "SetAnnotations",
+        "Params"  => {
+            "Tag"   => sprintf("%s-%s", $entity_type, $entity =~ s|/|-|g),
+            "Pairs" => $annotation
+        }
+    };
 
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "SetAnnotations",
-            "Params"  => {
-                "Tag"   => sprintf("%s-%s", $entity_type, $entity =~ s|/|-|g),
-                "Pairs" => $annotation
-            }
-        },
-        sub { my $res = shift; return $res }
-    );
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
 
-sub get_annotation {
+sub get_annotations {
     my ($self, $entity, $entity_type) = @_;
-    my $cb     = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "GetAnnotations",
+        "Params"  => {
+            "Tag" => "Tag" =>
+              sprintf("%s-%s", $entity_type, $entity =~ s|/|-|g)
+        }
+    };
 
-    $self->call(
-        {   "Type"    => "Client",
-            "Request" => "GetAnnotations",
-            "Params" =>
-              {"Tag" => sprintf("%s-%s", $entity_type, $entity =~ s|/|-|g)}
-        },
-        sub { my $res = shift; return $res }
-    );
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
 }
 
+sub private_address {
+    my ($self, $target) = @_;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "PrivateAddress",
+        "Params"  => {"Target" => $target}
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+}
+
+sub public_address {
+    my ($self, $target) = @_;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "PublicAddress",
+        "Params"  => {"Target" => $target}
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+}
+
+sub service_set_yaml {
+    my ($self, $service, $yaml) = @_;
+    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my $params = {
+        "Type"    => "Client",
+        "Request" => "PublicAddress",
+        "Params"  => {
+            "ServiceName" => $service,
+            "Config"      => Dump($yaml)
+        }
+    };
+
+    # block
+    return $self->call($params) unless $cb;
+
+    # non-block
+    return $self->call($params, $cb);
+}
+
+
 1;
 
 __END__
@@ -696,7 +924,7 @@ Juju::Environment - Exposed juju api environment
 
 =head1 VERSION
 
-version 1.0
+version 1.5.1
 
 =head1 SYNOPSIS
 
@@ -732,211 +960,773 @@ Supported juju jobs
 
 =head1 METHODS
 
-=head2 query_cs ($charm)
-
-helper for querying charm store for charm details
-
-=head2 _prepare_constraints ($constraints)
+=head2 _prepare_constraints
 
 Makes sure cpu-cores, cpu-power, mem are integers
 
-C<constraints> - hash of service constraints
+B<Params>
 
-B<Returns> - an updated constraint hash with any integers set properly.
+=over 4
+
+=item *
+
+C<constraints>
+
+hash of service constraints
+
+=back
 
 =head2 login
 
-Login to juju
+Login to juju, will die on a failed login attempt.
 
 =head2 reconnect
 
-Reconnects to API server in case of timeout
-
-=head2 info
-
-(Deprecated) Environment information
-
-B<Returns> - Juju environment state
+Reconnects to API server in case of timeout, this also resets the RequestId.
 
 =head2 environment_info
 
 Return Juju Environment information
 
+=head2 environment_uuid
+
+Environment UUID from client connection
+
+=head2 environment_unset
+
+Environment UUID from client connection
+
+B<Params>
+
+=over 4
+
+=item *
+
+C<items>
+
+=back
+
+=head2 find_tools
+
+Returns list containing all tools matching specified parameters
+
+B<Params>
+
+=over 4
+
+=item *
+
+C<major_verison>
+
+major version int
+
+=item *
+
+C<minor_verison>
+
+minor version int
+
+=item *
+
+C<series>
+
+Distribution series (eg, trusty)
+
+=item *
+
+C<arch>
+
+architecture
+
+=back
+
+=head2 agent_version
+
+Returns version of api server
+
+=head2 abort_current_upgrade
+
+Aborts and archives the current upgrade synchronization record, if any.
+
 =head2 status
 
 Returns juju environment status
 
+=head2 client_api_host_ports
+
+Returns network hostports for each api server
+
 =head2 get_watcher
 
 Returns watcher
 
-=head2 get_watched_tasks ($watcher_id)
+=head2 get_watched_tasks
 
 List of all watches for Id
 
-=head2 add_charm ($charm_url)
+B<Params>
+
+=over 4
+
+C<watcher_id>
+
+=back
+
+=head2 add_charm
 
 Add charm
 
-C<charm_url> - url of charm
+B<Params>
 
-=head2 get_charm ($charm_url)
+=over 4
+
+=item *
+
+C<charm_url>
+
+url of charm
+
+=back
+
+=head2 get_charm
 
 Get charm
 
-C<charm_url> - url of charm
+B<Params>
 
-=head2 get_env_constraints
+=over 4
+
+=item *
+
+C<charm_url>
+
+url of charm
+
+=back
+
+=head2 get_environment_constraints
 
 Get environment constraints
 
-=head2 set_env_constraints ($constraints)
+=head2 set_environment_constraints
 
 Set environment constraints
 
-C<constraints> - environment constraints
+B<Params>
 
-=head2 get_env_config
+=over 4
 
-=head2 set_env_config ($config)
+=item *
 
-C<config> - Config parameters
+C<constraints>
 
-=head2 add_machine ($series, $constraints, $machine_spec, $parent_id, $container_type)
+environment constraints
+
+=back
+
+=head2 environment_get
+
+Returns all environment settings
+
+=head2 environment_set
+
+Sets the given key-value pairs in the environment.
+
+B<Params>
+
+=over 4
+
+=item *
+
+C<config>
+
+Config parameters
+
+=back
+
+=head2 add_machine
 
 Allocate new machine from the iaas provider (i.e. MAAS)
 
-C<series> - OS series (i.e precise)
+B<Params>
 
-C<constraints> - machine constraints
+=over 4
 
-C<machine_spec> - not sure yet..
+=item *
 
-C<parent_id> - not sure yet..
+C<series>
 
-C<container_type> - uh..
+OS series (i.e precise)
 
-Note: Not quite right as I've no idea wtf its doing yet, need to read
-the specs.
+=item *
 
-=head2 add_machines ($machines)
+C<constraints>
+
+machine constraints
+
+=item *
+
+C<machine_spec>
+
+specific machine
+
+=item *
+
+C<parent_id>
+
+id of parent
+
+=item *
+
+C<container_type>
+
+kvm or lxc container type
+
+=back
+
+=head2 add_machines
 
 Add multiple machines from iaas provider
 
-C<machines> - List of machines
+B<Params>
+
+=over 4
+
+=item *
+
+C<machines>
+
+List of machines
+
+=back
+
+=head2 destroy_environment
+
+Destroys Juju environment
 
 =head2 destroy_machines
 
 Destroy machines
 
+B<Params>
+
+=over 4
+
+=item *
+
+C<machine_ids>
+
+List of machines
+
+=item *
+
+C<force>
+
+Force destroy
+
+=back
+
 =head2 provisioning_script
 
 Not implemented
 
-=head2 add_relation ($endpoint_a, $endpoint_b)
+=head2 add_relation
 
 Sets a relation between units
 
-=head2 remove_relation ($endpoint_a, $endpoint_b)
+B<Params>
+
+=over 4
+
+=item *
+
+C<endpoint_a>
+
+First unit endpoint
+
+=item *
+
+C<endpoint_b>
+
+Second unit endpoint
+
+=back
+
+=head2 remove_relation
 
 Removes relation between endpoints
 
-=head2 deploy ($service_name, $charm_url, $num_units, $config_yaml, $constraints, $machine_spec)
+B<Params>
+
+=over 4
+
+=item *
+
+C<endpoint_a>
+
+First unit endpoint
+
+=item *
+
+C<endpoint_b>
+
+Second unit endpoint
+
+=back
+
+=head2 deploy
 
 Deploys a charm to service
 
-=head2 set_config ($service_name, $config)
+B<Params>
+
+=over 4
+
+=item *
+
+C<charm>
+
+charm to deploy, can be in the format of B<series/charm> if needing to specify a different series
+
+=item *
+
+C<service_name>
+
+name of service to set. can be same name as charm, however, recommended to pick something unique and identifiable.
+
+=item *
+
+C<num_units>
+
+(optional) number of service units
+
+=item *
+
+C<config_yaml>
+
+(optional) A YAML formatted string of charm options
+
+=item *
+
+C<constraints>
+
+(optional) Machine hardware constraints
+
+=item *
+
+C<machine_spec>
+
+(optional) Machine specification
+
+=back
+
+More information on deploying can be found by running C<juju help deploy>.
+
+=head2 service_set
 
 Set's configuration parameters for unit
 
-C<service_name> - name of service (ie. blog)
+B<Params>
 
-C<config> - hash of config parameters
+=over 4
 
-=head2 unset_config ($service_name, $config_keys)
+=item *
+
+C<service_name>
+
+name of service (ie. blog)
+
+=item *
+
+C<config>
+
+hash of config parameters
+
+=back
+
+=head2 service_unset
 
 Unsets configuration value for service to restore charm defaults
 
-C<service_name> - name of service
+B<Params>
 
-C<config_keys> - hash of config keys to unset
+=over 4
 
-=head2 set_charm ($service_name, $charm_url, $force)
+=item *
+
+C<service_name>
+
+name of service
+
+=item *
+
+C<config_keys>
+
+config items to unset
+
+=back
+
+=head2 service_set_charm
 
 Sets charm url for service
 
-C<service_name> - name of service
+B<Params>
 
-C<charm_url> - charm location (ie. cs:precise/wordpress)
+=over 4
 
-=head2 get_service ($service_name)
+=item *
+
+C<service_name>
+
+name of service
+
+=item *
+
+C<charm_url> 
+
+charm location (ie. cs:precise/wordpress)
+
+=item *
+
+C<force>
+
+(optional) for setting charm url, overrides any existing charm url already set.
+
+=back
+
+=head2 service_get
 
 Returns information on charm, config, constraints, service keys.
 
+B<Params>
+
+=over 4
+
+=item *
+
 C<service_name> - name of service
 
-B<Returns> - Hash of information on service
+=back
 
-=head2 get_config ($service_name)
+=head2 get_config
 
 Get service configuration
 
-C<service_name> - name of service
+B<Params>
 
-B<Returns> - Hash of service configuration
+=over 4
 
-=head2 get_constraints ($service_name)
+=item *
 
-C<service_name> - Name of service
+C<service_name>
 
-=head2 set_constraints ($service_name, $constraints)
+name of service
 
-C<service_name> - Name of service
+=back
 
-C<constraints> - Service constraints
+=head2 get_service_constraints
 
-=head2 update_service ($service_name, $charm_url, $force_charm_url, $min_units, $settings, $constraints)
+Returns the constraints for the given service.
 
-Update a service
+B<Params>
 
-=head2 destroy_service ($service_name)
+=over 4
+
+=item *
+
+C<service_name>
+
+Name of service
+
+=back
+
+=head2 set_service_constraints
+
+Specifies the constraints for the given service.
+
+B<Params>
+
+=over 4
+
+=item *
+
+C<service_name>
+
+Name of service
+
+=item *
+
+C<constraints>
+
+Service constraints
+
+=back
+
+=head2 share_environment
+
+Allows the given users access to the environment.
+
+B<Params>
+
+=over 4
+
+=item *
+
+C<users>
+
+List of users to allow access
+
+=back
+
+=head2 unshare_environment
+
+Removes the given users access to the environment.
+
+B<Params>
+
+=over 4
+
+=item *
+
+C<users>
+
+List of users to remove access
+
+=back
+
+=head2 service_destroy
 
 Destroys a service
 
-C<service_name> - name of service
+B<Params>
 
-=head2 expose ($service_name)
+=over 4
+
+=item *
+
+C<service_name>
+
+name of service
+
+=back
+
+=head2 service_expose
 
 Expose service
 
-C<service_name> - Name of service
+B<Params>
 
-=head2 unexpose ($service_name)
+=over 4
+
+=item *
+
+C<service_name>
+
+Name of service
+
+=back
+
+=head2 service_unexpose
 
 Unexpose service
 
-C<service_name> - Name of service
+B<Params>
 
-=head2 valid_relation_names
+=over 4
+
+=item *
+
+C<service_name>
+
+Name of service
+
+=back
+
+=head2 service_charm_relations
 
 All possible relation names of a service
 
-=head2 add_units
+B<Params>
 
-=head2 add_unit
+=over 4
 
-=head2 remove_unit
+=item *
+
+C<service_name>
+
+Name of service
+
+=back
+
+=head2 add_service_units
+
+Adds given number of units to a service
+
+B<Params>
+
+=over 4
+
+=item *
+
+C<service_name>
+
+Name of service
+
+=item *
+
+C<num_units>
+
+Number of units to add
+
+=back
+
+=head2 add_service_unit
+
+Add unit to specific machine
+
+B<Params>
+
+=over 4
+
+=item *
+
+C<service_name>
+
+Name of service
+
+=item *
+
+C<machine_spec>
+
+Machine to add unit to
+
+=back
+
+=head2 destroy_service_units
+
+Decreases number of units dedicated to a service
+
+B<Params>
+
+=over 4
+
+=item *
+
+C<unit_names>
+
+List of units to destroy
+
+=back
 
 =head2 resolved
 
-=head2 get_public_address
+Clear errors on unit
 
-=head2 set_annotation
+B<Params>
+
+=over 4
+
+=item *
+
+C<unit_name>
+
+id of unit (eg, mysql/0)
+
+=item *
+
+C<retry>
+
+Boolean to force a retry
+
+=back
+
+=head2 run
+
+Not implemented yet.
+
+=head2 set_annotations
 
 Set annotations on entity, valid types are C<service>, C<unit>,
 C<machine>, C<environment>
 
-=head2 get_annotation
+=head2 get_annotations
+
+Returns annotations that have been set on the given entity.
+
+B<Params>
+
+=over 4
+
+=item *
+
+C<entity>
+
+=item *
+
+C<entity_type>
+
+=back
+
+=head2 private_address
+
+Get private address of machine or unit
+
+  $self->private_address('1');  # get address of machine 1
+  $self->private_address('mysql/0');  # get address of first unit of mysql
+
+B<Params>
+
+=over 4
+
+=item *
+
+C<target>
+
+Target machine
+
+=back
+
+=head2 public_address
+
+Returns the public address of the specified machine or unit. For a
+machine, target is an id not a tag.
+
+  $self->public_address('1');  # get address of machine 1
+  $self->public_address('mysql/0');  # get address of first unit of mysql
+
+B<Params>
+
+=over 4
+
+=item *
+
+C<target>
+
+Target machine
+
+=back
+
+=head2 service_set_yaml
+
+Sets configuration options on a service given options in YAML format.
+
+B<Params>
+
+=over 4
+
+=item *
+
+C<service>
+
+Service Name
+
+=item *
+
+C<yaml>
+
+YAML formatted string of options
+
+=back
 
 =head1 AUTHOR
 
@@ -950,6 +1740,18 @@ This is free software, licensed under:
 
   The MIT (X11) License
 
+=head1 SEE ALSO
+
+Please see those modules/websites for more information related to this module.
+
+=over 4
+
+=item *
+
+L<Juju|Juju>
+
+=back
+
 =head1 DISCLAIMER OF WARRANTY
 
 BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
